import ollama, os, argparse, threading, shutil, json
from ollama import Options, pull
from freegenius.utils.download import Downloader
from freegenius import config, is_valid_image_file, getOllamaServerClient
from freegenius import print1, print2, print3, toggleinputaudio, toggleoutputaudio
from freegenius.utils.ollama_models import ollama_models
from freegenius.utils.streaming_word_wrapper import StreamingWordWrapper
from freegenius.utils.single_prompt import SinglePrompt
from freegenius.utils.tool_plugins import Plugins

from prompt_toolkit.styles import Style
from prompt_toolkit import PromptSession
from prompt_toolkit.history import FileHistory
from prompt_toolkit.shortcuts import clear
from prompt_toolkit.completion import WordCompleter, FuzzyCompleter
#from prompt_toolkit.shortcuts import ProgressBar
from tqdm import tqdm
from pathlib import Path

promptStyle = Style.from_dict({
    # User input (default text).
    "": config.terminalCommandEntryColor2,
    # Prompt.
    "indicator": config.terminalPromptIndicatorColor2,
})

class OllamaChat:

    def __init__(self):
        # authentication
        if shutil.which("ollama"):
            self.runnable = True
        else:
            print("Local LLM Server 'Ollama' not found! Install Ollama first!")
            print("Visit https://ollama.com/")
            self.runnable = False

    def installModel(self, model):
        print3(f"Downloading '{model}' ...")
        
        #https://github.com/ollama/ollama-python/blob/main/examples/pull-progress/main.py
        current_digest, bars = '', {}
        for progress in pull(model, stream=True):
            digest = progress.get('digest', '')
            if digest != current_digest and current_digest in bars:
                bars[current_digest].close()

            if not digest:
                print(progress.get('status'))
                continue

            if digest not in bars and (total := progress.get('total')):
                bars[digest] = tqdm(total=total, desc=f'pulling {digest[7:19]}', unit='B', unit_scale=True)

            if completed := progress.get('completed'):
                bars[digest].update(completed - bars[digest].n)

            current_digest = digest

    def resetMessages(self):
        return [{"role": "system", "content": config.systemMessage_ollama}]

    def setSystemMessage(self):
        # completer
        Plugins.runPlugins()
        completer = FuzzyCompleter(WordCompleter(list(config.predefinedContexts.values()), ignore_case=True))
        # history
        historyFolder = os.path.join(config.localStorage, "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        system_message_history = os.path.join(historyFolder, "system_message")
        system_message_session = PromptSession(history=FileHistory(system_message_history))
        # prompt
        print2("Change system message below:")
        prompt = SinglePrompt.run(style=promptStyle, promptSession=system_message_session, default=config.systemMessage_ollama, completer=completer)
        if prompt and not prompt == config.exit_entry:
            config.systemMessage_ollama = prompt
            config.saveConfig()
            print2("System message changed!")
            clear()
            self.messages = self.resetMessages()
            print("New chat started!")

    def run(self, prompt="", model="mistral", once=False) -> None:        
        def extractImages(content) -> list:
            template = {
                "image_filepath_list": [],
                "query": "",
            }
            promptPrefix = f"""Use this template:

{template}

To generate a JSON that contains two keys, "image_filepath_list" and "query", based on my request.
"image_filepath_list" is a list of image paths specified in my request.  If no path is specified, return an empty list [] for its value.
"query" is the query about the images in the list.

Here is my request:
"""
            completion = getOllamaServerClient().chat(
                model="gemma:2b",
                messages=[
                    {
                        "role": "user",
                        "content": f"{promptPrefix}{content}",
                    },
                ],
                format="json",
                stream=False,
                options=Options(
                    temperature=0.0,
                    num_ctx=config.ollamaMainModel_num_ctx,
                    num_batch=config.ollamaMainModel_num_batch,
                    num_predict=config.ollamaMainModel_num_predict,
                ),
            )
            output = json.loads(completion["message"]["content"])
            if config.developer:
                print2("Input:")
                print(output)
            imageList = output["image_filepath_list"]
            images = [i for i in imageList if os.path.isfile(i) and is_valid_image_file(i)]

            return images

        if not self.runnable:
            return None

        # check model
        if not Downloader.downloadOllamaModel(model):
            return None
        if model.startswith("llava"):
            Downloader.downloadOllamaModel("gemma:2b")
        
        previoiusModel = config.ollamaMainModel
        config.ollamaMainModel = model
        if not config.ollamaMainModel:
            config.ollamaMainModel = "mistral"
        if not config.ollamaMainModel == previoiusModel:
            config.saveConfig()

        historyFolder = os.path.join(config.localStorage, "history")
        Path(historyFolder).mkdir(parents=True, exist_ok=True)
        chat_history = os.path.join(historyFolder, f"ollama_{model}")
        chat_session = PromptSession(history=FileHistory(chat_history))

        print2(f"\n{model.capitalize()} loaded!")
        print2("```system message")
        print1(config.systemMessage_ollama)
        print2("```")

        # history
        if hasattr(config, "currentMessages"):
            messages = []
            for i in config.currentMessages[:-1]:
                if "role" in i and i["role"] in ("system", "user", "assistant") and "content" in i and i.get("content"):
                    messages.append(i)
        else:
            messages = self.resetMessages()

        # bottom toolbar
        if hasattr(config, "currentMessages"):
            bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry}"""
        else:
            bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry} {str(config.hotkey_new).replace("'", "")} .new"""
            print("(To start a new chart, enter '.new')")
        print(f"(To exit, enter '{config.exit_entry}')\n")

        while True:
            completer = None if hasattr(config, "currentMessages") else FuzzyCompleter(WordCompleter([".new", ".systemmessage", ".toggleinputaudio", ".toggleoutputaudio", config.exit_entry], ignore_case=True))
            if not prompt:
                prompt = SinglePrompt.run(style=promptStyle, promptSession=chat_session, bottom_toolbar=bottom_toolbar, completer=completer)
                if prompt and not prompt in (".new", config.exit_entry) and hasattr(config, "currentMessages"):
                    config.currentMessages.append({"content": prompt, "role": "user"})
            else:
                prompt = SinglePrompt.run(style=promptStyle, promptSession=chat_session, bottom_toolbar=bottom_toolbar, default=prompt, accept_default=True, completer=completer)
            if prompt == config.exit_entry:
                break
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".toggleinputaudio":
                toggleinputaudio()
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".toggleoutputaudio":
                toggleoutputaudio()
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".systemmessage":
                self.setSystemMessage()
            elif not hasattr(config, "currentMessages") and prompt.lower() == ".new":
                clear()
                messages = self.resetMessages()
                print("New chat started!")
            elif prompt := prompt.strip():
                streamingWordWrapper = StreamingWordWrapper()
                config.pagerContent = ""
                if model.startswith("llava"):
                    images = extractImages(prompt)
                    if images:
                        messages.append({'role': 'user', 'content': prompt, 'images': images})
                        print3(f"Analyzing image: {str(images)}")
                    else:
                        messages.append({'role': 'user', 'content': prompt})
                else:
                    messages.append({'role': 'user', 'content': prompt})
                try:
                    completion = getOllamaServerClient().chat(
                        model=model,
                        messages=messages,
                        stream=True,
                        options=Options(
                            temperature=config.llmTemperature,
                            num_ctx=config.ollamaMainModel_num_ctx,
                            num_batch=config.ollamaMainModel_num_batch,
                            num_predict=config.ollamaMainModel_num_predict,
                        ),
                    )
                    # Create a new thread for the streaming task
                    streaming_event = threading.Event()
                    self.streaming_thread = threading.Thread(target=streamingWordWrapper.streamOutputs, args=(streaming_event, completion,))
                    # Start the streaming thread
                    self.streaming_thread.start()

                    # wait while text output is steaming; capture key combo 'ctrl+q' or 'ctrl+z' to stop the streaming
                    streamingWordWrapper.keyToStopStreaming(streaming_event)

                    # when streaming is done or when user press "ctrl+q"
                    self.streaming_thread.join()

                    # update messages
                    messages.append({"role": "assistant", "content": config.new_chat_response})
                except ollama.ResponseError as e:
                    if hasattr(self, "streaming_thread"):
                        self.streaming_thread.join()
                    print('Error:', e.error)

            prompt = ""

            if once:
                break

        print2(f"\n{model.capitalize()} closed!")
        if hasattr(config, "currentMessages"):
            print2(f"Return back to {config.freeGeniusAIName} prompt ...")

# available cli: 'ollamachat', 'mistral', 'llama3', 'llama370b', 'gemma2b', 'gemma7b', 'llava', 'phi3', 'vicuna', 'wizardlm2'

def wizardlm2():
    main("wizardlm2")

def starlinglm():
    main("starling-lm")

def orca2():
    main("orca2")

def mistral():
    main("mistral")

def mixtral():
    main("mixtral")

def llama3():
    main("llama3")

def llama370b():
    main("llama3:70b")

def codellama():
    main("codellama")

def gemma2b():
    main("gemma:2b")

def gemma7b():
    main("gemma:7b")

def llava():
    main("llava")

def phi3():
    main("phi3")

def vicuna():
    main("vicuna")

def main(thisModel=""):
    # Create the parser
    parser = argparse.ArgumentParser(description="palm2 cli options")
    # Add arguments
    parser.add_argument("default", nargs="?", default=None, help="default entry")
    if not thisModel:
        parser.add_argument('-m', '--model', action='store', dest='model', help="specify language model with -m flag; default: mistral")
    # Parse arguments
    args = parser.parse_args()
    # Get options
    prompt = args.default.strip() if args.default and args.default.strip() else ""
    if thisModel:
        model = thisModel
    else:
        if args.model and args.model.strip():
            model = args.model.strip()
        else:
            historyFolder = os.path.join(config.localStorage, "history")
            Path(historyFolder).mkdir(parents=True, exist_ok=True)
            model_history = os.path.join(historyFolder, "ollama_default")
            model_session = PromptSession(history=FileHistory(model_history))
            completer = FuzzyCompleter(WordCompleter(sorted(ollama_models), ignore_case=True))
            bottom_toolbar = f""" {str(config.hotkey_exit).replace("'", "")} {config.exit_entry}"""

            print2("Ollama chat launched!")
            print("Select a model below:")
            print("Note: You should have at least 8 GB of RAM available to run the 7B models, 16 GB to run the 13B models, and 32 GB to run the 33B models.")
            model = SinglePrompt.run(style=promptStyle, promptSession=model_session, bottom_toolbar=bottom_toolbar, default=config.ollamaMainModel, completer=completer)
            if model and model.lower() == config.exit_entry:
                print2("\nOllama chat closed!")
                return None

    if not model:
        model = config.ollamaMainModel
    # Run chat bot
    OllamaChat().run(
        prompt=prompt,
        model=model,
    )
        

if __name__ == '__main__':
    main()


import os
from freegenius import config, print2, print3, getUnstructuredFiles, get_or_create_collection, add_vector, query_vectors
from freegenius.utils.call_llm import CallLLM
import os, traceback, re, zipfile, datetime, traceback, threading, shutil, chromadb
from chromadb.config import Settings
from pathlib import Path
from freegenius.utils.prompts import Prompts
from prompt_toolkit import print_formatted_text, HTML
from prompt_toolkit.styles import Style
from autogen.retrieve_utils import TEXT_FORMATS
from langchain.text_splitter import RecursiveCharacterTextSplitter
from langchain_community.document_loaders import UnstructuredFileLoader
from langchain_openai import OpenAIEmbeddings
# from langchain_community.embeddings.sentence_transformer import SentenceTransformerEmbeddings
# from langchain_community.embeddings import HuggingFaceEmbeddings
from langchain_huggingface import HuggingFaceEmbeddings
from langchain_community.vectorstores import Chroma
from freegenius.utils.streaming_word_wrapper import StreamingWordWrapper


class RAG:

    def __init__(self):
        CallLLM.checkCompletion()

        rag_store = os.path.join(config.localStorage, "rag")
        try:
            shutil.rmtree(rag_store, ignore_errors=True)
            #print2("Old retrieval store removed!")
        except:
            print2("Failed to remove old rag store!")
        Path(rag_store).mkdir(parents=True, exist_ok=True)
        chroma_client = chromadb.PersistentClient(rag_store, Settings(anonymized_telemetry=False))
        self.collection = get_or_create_collection(chroma_client, "rag")

    def getContext1(self, splits, query):
        for i in splits:
            i.metadata["source"] = i.metadata["source"][0]
        # https://python.langchain.com/docs/integrations/text_embedding/sentence_transformers
        #embeddings = HuggingFaceEmbeddings(model_name="all-MiniLM-L6-v2")
        #embedding = OpenAIEmbeddings(model=config.embeddingModel) if config.embeddingModel in ("text-embedding-3-large", "text-embedding-3-small", "text-embedding-ada-002") else SentenceTransformerEmbeddings(model_name=config.embeddingModel)
        embedding = OpenAIEmbeddings(model=config.embeddingModel) if config.embeddingModel in ("text-embedding-3-large", "text-embedding-3-small", "text-embedding-ada-002") else HuggingFaceEmbeddings(model_name=config.embeddingModel)
        # https://python.langchain.com/docs/integrations/vectorstores/chroma
        # https://github.com/langchain-ai/langchain/issues/7804
        vectorstore = Chroma.from_documents(
            documents=splits,
            embedding=embedding,
            client_settings=Settings(anonymized_telemetry=False),
        )
        # reference: https://python.langchain.com/docs/modules/data_connection/retrievers/vectorstore
        # Create the retriever
        #retriever_settings = {"search_type": "similarity_score_threshold", "search_kwargs": {"score_threshold": 0.5}}
        #retriever_settings = {"search_type": "mmr"}
        if config.rag_retrieverSettings: # default: {'search_kwargs': {'k': 5}}
            # align setting with config.rag_closestMatches
            if not "search_kwargs" in config.rag_retrieverSettings:
                config.rag_retrieverSettings["search_kwargs"] = {"k": 5}
            elif "search_kwargs" in config.rag_retrieverSettings and (not "k" in config.rag_retrieverSettings["search_kwargs"] or ("k" in config.rag_retrieverSettings["search_kwargs"] and not config.rag_retrieverSettings["search_kwargs"]["k"] == config.rag_closestMatches)):
                    config.rag_retrieverSettings["search_kwargs"]["k"] = config.rag_closestMatches
            retriever = vectorstore.as_retriever(**config.rag_retrieverSettings)
        else:
            retriever = vectorstore.as_retriever()
        # retrieve document
        retrieved_docs = retriever.invoke(query)
        if not retrieved_docs:
            return None
        else:
            formatted_context = {f"information_{index}": item.page_content for index, item in enumerate(retrieved_docs)}
        return formatted_context

    def getContext2(self, splits, query):
        # add to vector database
        for i in splits:
            add_vector(self.collection, i.page_content, metadata={"source": i.metadata.get("source")[0]})
        
        # search
        retrieved_docs = query_vectors(self.collection, query, n=config.rag_closestMatches)["documents"]
        if not retrieved_docs:
            return None

        # format context
        #formatted_context = "\n\n".join([i.page_content for i in retrieved_docs])
        formatted_context = {f"information_{index}": content for index, content in enumerate(retrieved_docs)}
        #formatted_context = [i[0] for i in retrieved_docs]
        return formatted_context

    def getResponse(self, docs_path, query):
        if not os.path.exists(docs_path):
            print2("Invalid path!")
            return None

        rag = os.path.join(config.localStorage, "rag")
        Path(rag).mkdir(parents=True, exist_ok=True)

        _, file_extension = os.path.splitext(docs_path)
        if file_extension.lower() == ".zip":
            # support zip file; unzip zip file, if any
            currentTime = re.sub("[\. :]", "_", str(datetime.datetime.now()))
            extract_to_path = os.path.join(rag, "unpacked", currentTime)
            print3(f"Unpacking content to: {extract_to_path}")
            if not os.path.isdir(extract_to_path):
                Path(rag).mkdir(parents=True, exist_ok=True)
            with zipfile.ZipFile(docs_path) as zip_ref:
                zip_ref.extractall(extract_to_path)
            docs_path = extract_to_path
        # check if file format is supported
        if os.path.isfile(docs_path):
            if file_extension[1:] in TEXT_FORMATS:
                docs_path = [docs_path]
            else:
                print2("File format not supported!")
                return None
        elif os.path.isdir(docs_path):
            docs_path = getUnstructuredFiles(docs_path)
            if not docs_path:
                print2("Support files not found!")
                return None
        else:
            print2("Document path invalid!")
            return None


        # https://python.langchain.com/docs/integrations/providers/unstructured
        loader = UnstructuredFileLoader(docs_path) # file_path: Union[str, List[str]]
        doc = loader.load()
        #print(doc)

        #chunk it
        text_splitter = RecursiveCharacterTextSplitter(chunk_size=1000, chunk_overlap=200)
        splits = text_splitter.split_documents(doc)
        #print(splits)
        formatted_context = self.getContext1(splits, query)
        if not formatted_context:
            with open(docs_path[0], "r", encoding="utf-8") as fileObj:
                formatted_context = fileObj.read()

        # format prompt
        formatted_prompt = f"""Question:
<question>
{query}
</question>

Context:
<context>
{formatted_context}
</context>

Please answer my question, based on the context given above."""

        if hasattr(config, "currentMessages"):
            messages = config.currentMessages[:-1] + [{"role": "user", "content" : formatted_prompt}]
        else:
            messages = [{"role": "user", "content" : formatted_prompt}]

        streamingWordWrapper = StreamingWordWrapper()
        config.pagerContent = ""

        try:
            completion = CallLLM.regularCall(messages)

            # Create a new thread for the streaming task
            streaming_event = threading.Event()
            self.streaming_thread = threading.Thread(target=streamingWordWrapper.streamOutputs, args=(streaming_event, completion, True if config.llmInterface in ("chatgpt", "letmedoit", "groq", "llamacppserver") else False))
            # Start the streaming thread
            self.streaming_thread.start()

            # wait while text output is steaming; capture key combo 'ctrl+q' or 'ctrl+z' to stop the streaming
            streamingWordWrapper.keyToStopStreaming(streaming_event)

            # when streaming is done or when user press "ctrl+q"
            self.streaming_thread.join()
        except:
            print(traceback.format_exc())
            self.streaming_thread.join()
        
        return ""

    def print(self, message):
        #print(message)
        print_formatted_text(HTML(message))

    def refinePath(self, docs_path):
        docs_path = docs_path.strip()
        docs_path = re.sub("^'(.*?)'$", r"\1", docs_path)
        if "\\ " in docs_path or "\(" in docs_path:
            docs_path = docs_path.replace("\\ ", " ")
            docs_path = docs_path.replace("\(", "(")
        return os.path.expanduser(docs_path)

    def run(self):
        promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })
        prompts = Prompts()

        print2(f"Retrieval utility launched!")
        self.print(f"""[press '{str(config.hotkey_exit).replace("'", "")[1:-1]}' to exit]""")
        
        print2(f"Enter document path below (file / folder):")
        self.print(f"""Supported formats: *.{", *.".join(TEXT_FORMATS)}""" + ", *.zip")
        docs_path = prompts.simplePrompt(style=promptStyle)

        # handle path dragged to terminal
        docs_path = self.refinePath(docs_path)

        if docs_path and os.path.exists(docs_path):
            self.print("Enter your query below:")
            query = prompts.simplePrompt(style=promptStyle)
            if not query == config.exit_entry:
                try:
                    self.getResponse(docs_path, query)
                except:
                    print(traceback.format_exc())
        else:
            print2(f"Entered path does not exist!")
        
        print2(f"Retrieval utility closed!")

def main():
    RAG().run()

if __name__ == '__main__':
    main()


"""
FreeGenius AI Plugin - input suggestions

add input suggestions
"""

from freegenius import config
import sys, os

config.inputSuggestions += [
        "@command ",
        "@append_command ",
        #"[CHAT_chatgpt] ",
        #"[CHAT_geminipro] ",
        #"[CHAT_palm2] ",
        #"[CHAT_codey] ",
        f"!{config.open} ",
        f"!{sys.executable} ",
        "open with default application: ",
        "open with file manager: ",
        "open with web browser: ",
        "read ",
        "search ",
        "analyze ",
        "tell me about ",
        "write a summary ",
        "explain ",
        "What does it mean? ",
        "@chat Craft a prompt for ChatGPT that outlines the necessary steps it should take to complete the following task at hand:\n",
        f"@chat Improve the following content according to {config.improvedWritingSytle}:\n",
        "Before you start, please ask me any questions you have about this so I can give you more context. Be extremely comprehensive.",
        "Write a short and complete summary of the text below:\n",
        "Ues the provided text and list 5 main topics.",
        "Using the provided text, raise 5 short questions that can help investigating the subject. The questions must be of the type 'who', 'what', 'when', 'how', 'why'.",
        """You are a Language Model trained to answer questions based solely on the provided text. The provided text is your only source of information, and you must not use any external knowledge. If you do not have enough information from the text to answer a question, please reply with 'I don't know.'

Question: {question}

```
{text}
```""",
        """Given a block of text, please extract and return the list of all sections and main topics. The list must provide a clear and concise overview of what is being discussed. This list must help readers to quickly find the information they need and navigate through the content. Avoid repetitions.

```
{text}
```

Index of the topics:""",
    ]

config.inputSuggestions.append("""@chat Translate Content. Assist me by acting as a translator. Once I have provided you with the content, you should inquire about the language I need it translated into. After I inform you of the desired language, proceed with the translation.

Please translate the content below:
""")


import vertexai, os, argparse, re
from vertexai.generative_models import GenerativeModel
from vertexai.generative_models._generative_models import (
    GenerationConfig,
    HarmCategory,
    HarmBlockThreshold,
)
from freegenius import config, showErrors, is_valid_image_file, is_valid_image_url, wrapText, print2, print3, is_valid_url, toggleinputaudio, toggleoutputaudio
from freegenius.utils.single_prompt import SinglePrompt
from freegenius.utils.tool_plugins import Plugins

import http.client
import typing
import urllib.request
from vertexai.generative_models import Image
from prompt_toolkit.styles import Style
from prompt_toolkit.completion import WordCompleter, FuzzyCompleter


# Install google-cloud-aiplatform FIRST!
#!pip install --upgrade google-cloud-aiplatform


class GeminiProVision:

    def __init__(self, temperature=0.9, max_output_tokens=2048):
        # authentication
        if os.environ["GOOGLE_APPLICATION_CREDENTIALS"] and "Vertex AI" in config.enabledGoogleAPIs:
            self.runnable = True
        else:
            print("Vertex AI is disabled!")
            print("Read https://github.com/eliranwong/letmedoit/wiki/Google-API-Setup for setting up Google API.")
            self.runnable = False
        # initiation
        vertexai.init()
        self.generation_config=GenerationConfig(
            temperature=temperature, # 0.0-1.0; default 0.9
            max_output_tokens=max_output_tokens, # default
            candidate_count=1,
        )
        self.safety_settings={
            HarmCategory.HARM_CATEGORY_UNSPECIFIED: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_DANGEROUS_CONTENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_HATE_SPEECH: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_HARASSMENT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
            HarmCategory.HARM_CATEGORY_SEXUALLY_EXPLICIT: HarmBlockThreshold.BLOCK_ONLY_HIGH,
        }
        self.promptStyle = Style.from_dict({
            # User input (default text).
            "": config.terminalCommandEntryColor2,
            # Prompt.
            "indicator": config.terminalPromptIndicatorColor2,
        })

    def refinePath(self, img_path):
        img_path = img_path.strip()
        img_path = re.sub("^'(.*?)'$", r"\1", img_path)
        if "\\ " in img_path or "\(" in img_path:
            img_path = img_path.replace("\\ ", " ")
            img_path = img_path.replace("\(", "(")
        return os.path.expanduser(img_path)

    def run(self, query="", files=[]):
        print2("\nGemini Pro Vision loaded!")
        print(f"""[press '{str(config.hotkey_exit).replace("'", "")[1:-1]}' to exit]""")
        if not files:
            print2("Enter image path below (file / folder):")
            files = SinglePrompt.run(style=self.promptStyle)
        if files:
            # handle path dragged to terminal
            files = self.refinePath(files)
        if files == config.exit_entry:
            pass
        elif files and os.path.exists(files):
            files = [files]
            if not query:
                print2("Enter your query below:")
                query = SinglePrompt.run(style=self.promptStyle)
            if query and not query == config.exit_entry:
                try:
                    function_args = {
                        "query": query,
                        "image_filepath": files,
                    }
                    self.analyze_images(function_args)
                except:
                    showErrors()
        else:
            print2("Entered path does not exist!")
        print2("\nGemini Pro Vision closed!")

    def analyze_images(self, function_args):
        def load_image_from_url(image_url: str) -> Image:
            with urllib.request.urlopen(image_url) as response:
                response = typing.cast(http.client.HTTPResponse, response)
                image_bytes = response.read()
            return Image.from_bytes(image_bytes)
        def load_image_from_file(file_path: str) -> Image:
            with open(file_path, "rb") as file:
                image_bytes = file.read()
            return Image.from_bytes(image_bytes)

        query = function_args.get("query") # required
        files = function_args.get("image_filepath") # required
        if not files:
            self.defaultPrompt = f"@chat {query}"
            return None
        if isinstance(files, str):
            if not files.startswith("["):
                files = f'["{files}"]'
            files = eval(files)
        else:
            files = list(files)

        # check if some items are directory that contain images
        filesCopy = files[:]
        for item in filesCopy:
            if os.path.isdir(item):
                for root, _, allfiles in os.walk(item):
                    for file in allfiles:
                        file_path = os.path.join(root, file)
                        files.append(file_path)
                files.remove(item)

        content = []
        imageFiles = []
        # validate image paths
        for i in files:
            if is_valid_url(i) and is_valid_image_url(i):
                content.append(load_image_from_url(i))
                imageFiles.append(i)
            elif os.path.isfile(i) and is_valid_image_file(i):
                content.append(load_image_from_file(i))
                imageFiles.append(i)
        content.append(f"{query} Please give your answer in as much detail as possible.")

        if imageFiles:
            print3(f"Reading: '{', '.join(imageFiles)}'")
            model = GenerativeModel("gemini-pro-vision")
            response = model.generate_content(
                content,
                generation_config=self.generation_config,
                safety_settings=self.safety_settings,
            )
            if response:
                try:
                    chat_response = response.text.strip()
                    if chat_response:
                        print2("```assistant")
                        print(wrapText(chat_response))
                        print2("```")
                        return chat_response
                except:
                    showErrors()
                    return ""
        else:
            print2("No image file is found!")
        return ""

def main():
    # Create the parser
    parser = argparse.ArgumentParser(description="geminipro vision cli options")
    # Add arguments
    parser.add_argument("default", nargs="?", default=None, help="default entry")
    parser.add_argument('-f', '--files', action='store', dest='files', help="specify a file of a directory of files with -f flag")
    parser.add_argument('-o', '--outputtokens', action='store', dest='outputtokens', help="specify maximum output tokens with -o flag; default: 2048")
    parser.add_argument('-t', '--temperature', action='store', dest='temperature', help="specify temperature with -t flag: default: 0.9")
    # Parse arguments
    args = parser.parse_args()
    # Get options
    query = args.default.strip() if args.default and args.default.strip() else ""
    files = [args.files.strip()] if args.files and os.path.exists(args.files.strip()) else []
    if args.outputtokens and args.outputtokens.strip():
        try:
            max_output_tokens = int(args.outputtokens.strip())
        except:
            max_output_tokens = 2048
    else:
        max_output_tokens = 2048
    if max_output_tokens > 2048:
        max_output_tokens = 2048
    if args.temperature and args.temperature.strip():
        try:
            temperature = float(args.temperature.strip())
        except:
            temperature = 0.9
    else:
        temperature = 0.9
    if temperature > 1.0:
        temperature = 1.0
    GeminiProVision(
        temperature=temperature,
        max_output_tokens = max_output_tokens,
    ).run(
        query=query,
        files=files,
    )

if __name__ == '__main__':
    main()


# install binary ffmpeg and python package yt-dlp to work with this plugin

"""
FreeGenius AI Plugin - download youtube or web content

* download Youtube video
* download Youtube audio and convert it into mp3
* download webcontent

[FUNCTION_CALL]
"""

from freegenius import config, showErrors, isCommandInstalled, print1, print3, is_valid_url, downloadWebContent
import re, subprocess, os
from pathlib import Path


def download_web_content(function_args):
    def is_youtube_url(url_string):
        pattern = r'(?:https?:\/\/)?(?:www\.)?youtu(?:\.be|be\.com)\/(?:watch\?v=|embed\/|v\/)?([a-zA-Z0-9\-_]+)'
        match = re.match(pattern, url_string)
        return match is not None

    def isFfmpegInstalled():
        ffmpegVersion = subprocess.Popen("ffmpeg -version", shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
        *_, stderr = ffmpegVersion.communicate()
        return False if stderr else True

    def terminalDownloadYoutubeFile(downloadCommand, url_string, outputFolder):
        if isFfmpegInstalled():
            try:
                print1("--------------------")
                # use os.system, as it displays download status ...
                os.system("cd {2}; {0} {1}".format(downloadCommand, url_string, outputFolder))
                if isCommandInstalled("pkill"):
                    os.system("pkill yt-dlp")
                print3(f"Downloaded in: '{outputFolder}'")
                try:
                    os.system(f'''{config.open} {outputFolder}''')
                except:
                    pass
            except:
                showErrors() 
        else:
            print1("Tool 'ffmpeg' is not found on your system!")
            print1("Read https://github.com/eliranwong/letmedoit/wiki/Install-ffmpeg")


    url = function_args.get("url") # required
    if is_youtube_url(url):
        print1("Loading youtube downloader ...")
        format = "video"
        location = function_args.get("location", "") # optional
        if not (location and os.path.isdir(location)):
            location = os.path.join(config.localStorage, "audio" if format == "audio" else "video")
            Path(location).mkdir(parents=True, exist_ok=True)
        downloadCommand = "yt-dlp -x --audio-format mp3" if format == "audio" else "yt-dlp -f bestvideo[ext=mp4]+bestaudio[ext=m4a]/mp4"
        terminalDownloadYoutubeFile(downloadCommand, url, location)
        return "Finished! Youtube downloader closed!"
    elif is_valid_url(url):
        try:
            folder = config.localStorage
            folder = os.path.join(folder, "Downloads")
            Path(folder).mkdir(parents=True, exist_ok=True)
            downloadWebContent(url, folder=folder, ignoreKind=True)
            return ""
        except:
            showErrors()
            return "[INVALID]"
    else:
        print1("invalid link given")
        return "[INVALID]"

functionSignature = {
    "examples": [
        "download Youtube video",
        "download Youtube audio into mp3 format",
        "download webpage",
    ],
    "name": "download_web_content",
    "description": "download Youtube video into mp4 file or download audio into mp3 file or download webcontent",
    "parameters": {
        "type": "object",
        "properties": {
            "url": {
                "type": "string",
                "description": "url given by user",
            },
            "location": {
                "type": "string",
                "description": "Output folder where downloaded file is to be saved",
            },
        },
        "required": ["url"],
    },
}

config.addFunctionCall(signature=functionSignature, method=download_web_content)


from freegenius import config, count_tokens_from_messages, count_tokens_from_functions, tokenLimits
from prompt_toolkit.validation import Validator, ValidationError
from prompt_toolkit.application import run_in_terminal
import tiktoken
import re

class TokenValidator(Validator):
    def validate(self, document):
        #current_buffer = get_app().current_buffer
        currentInput = document.text
        if not config.dynamicTokenCount or not currentInput or currentInput.lower() in (config.exit_entry, config.cancel_entry, ".new", ".share", ".save"):
            pass
        else:
            try:
                encoding = tiktoken.encoding_for_model(config.chatGPTApiModel)
            except:
                encoding = tiktoken.get_encoding("cl100k_base")
            ''' does not apply for multiple tool cases
            no_function_call_pattern = "\[NO_TOOL\]|\[CHAT\]|\[CHAT_[^\[\]]+?\]"
            #if "[NO_TOOL]" in currentInput:
            if re.search(no_function_call_pattern, currentInput):
                availableFunctionTokens = 0
                #currentInput = currentInput.replace("[NO_TOOL]", "")
                currentInput = re.sub(no_function_call_pattern, "", currentInput)
            else:
                availableFunctionTokens = count_tokens_from_functions(config.toolFunctionSchemas)
            '''
            availableFunctionTokens = count_tokens_from_functions(config.toolFunctionSchemas)
            currentInputTokens = len(encoding.encode(config.fineTuneUserInput(currentInput)))
            loadedMessageTokens = count_tokens_from_messages(config.currentMessages)
            selectedModelLimit = tokenLimits[config.chatGPTApiModel]
            #estimatedAvailableTokens = selectedModelLimit - availableFunctionTokens - loadedMessageTokens - currentInputTokens

            config.dynamicToolBarText = f""" Tokens: {(availableFunctionTokens + loadedMessageTokens + currentInputTokens)}/{selectedModelLimit} {str(config.hotkey_display_key_combo).replace("'", "")} shortcuts """
            #if config.conversationStarted:
            #    config.dynamicToolBarText = config.dynamicToolBarText + " [ctrl+n] new"
            if selectedModelLimit - (availableFunctionTokens + loadedMessageTokens + currentInputTokens) >= config.chatGPTApiMinTokens:
                pass
            else:
                run_in_terminal(lambda: print(f"""Press '{str(config.hotkey_new).replace("'", "")[1:-1]}' to start a new chat!"""))
                raise ValidationError(message='Token limit reached!', cursor_position=document.cursor_position)

class NumberValidator(Validator):
    def validate(self, document):
        text = document.text

        if text.lower() in (config.exit_entry, config.cancel_entry):
            pass
        elif text and not text.isdigit():
            i = 0

            # Get index of first non numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):
                if not c.isdigit():
                    break

            raise ValidationError(message='This entry accepts numbers only!', cursor_position=i)


class FloatValidator(Validator):
    def validate(self, document):
        text = document.text

        if text.lower() in (config.exit_entry, config.cancel_entry):
            pass
        try:
            float(text)
        except:
            raise ValidationError(message='This entry accepts floating point numbers only!', cursor_position=0)


class NoAlphaValidator(Validator):
    def validate(self, document):
        text = document.text

        if text.lower() in (config.exit_entry, config.cancel_entry):
            pass
        elif text and text.isalpha():
            i = 0

            # Get index of first non numeric character.
            # We want to move the cursor here.
            for i, c in enumerate(text):
                if c.isalpha():
                    break

            raise ValidationError(message='This entry does not accept alphabet letters!', cursor_position=i)


from freegenius import config, fineTunePythonCode
from freegenius import print1, print2, print3
import traceback
from freegenius import installPipPackage
from freegenius.utils.python_utils import PythonUtil

"""
FreeGenius AI Plugin - auto correct python code

functionalities:
* install missing packages
* fixed broken codes

User can define the maximum number of auto-correctioning attempts by editing "max_consecutive_auto_correction" in config.py.
The default value of config.max_consecutive_auto_correction is 3.

[FUNCTION_CALL]
"""

def correct_python(function_args):
    # get the sql query statement
    issue = function_args.get("issue") # required
    print(f"Issue: {issue}")

    fix = function_args.get("code") # required
    missing = function_args.get("missing_module") # required
    if missing in ("[]", "???"):
        missing = ""

    try:
        if missing:
            try:
                print3(f"Installing missing package: {missing}")
                installPipPackage(f"--upgrade {missing}")
            except:
                print(traceback.format_exc())
        print2("Running improved code ...")
        if config.developer or config.codeDisplay:
            PythonUtil.displayPythonCode(fix)
        exec(fineTunePythonCode(fix), globals())
        return "EXECUTED"
    except:
        return traceback.format_exc()

functionSignature = {
    "examples": [
        "fix bug",
        "correct code",
    ],
    "name": "correct_python",
    "description": "Fix python code if both original code and traceback error are provided",
    "parameters": {
        "type": "object",
        "properties": {
            "code": {
                "type": "string",
                "description": "Generate an improved version of python code that resolved the traceback error. Return the original code only if traceback shows an import error.",
            },
            "missing_module": {
                "type": "string",
                "description": """The module name identified in ModuleNotFoundError, if any. Return '' if there is no import error in the traceback.""",
            },
            "issue": {
                "type": "string",
                "description": """Briefly explain the error""",
            },
        },
        "required": ["code", "missing_module", "issue"],
    },
}

# configs particular to this plugin
# persistent
persistentConfigs = (
    ("max_consecutive_auto_correction", 5),
)
config.setConfig(persistentConfigs)

config.addFunctionCall(signature=functionSignature, method=correct_python)